/****************************************************************************
**
** Copyright (C) 2023 The Qt Company Ltd.
** Contact: https://www.qt.io/licensing/
**
** This file is part of the Qt Design Viewer of the Qt Toolkit.
**
** Commercial License Usage
** Licensees holding valid commercial Qt licenses may use this file in
** accordance with the commercial license agreement provided with the
** Software or, alternatively, in accordance with the terms contained in
** a written agreement between you and The Qt Company. For licensing terms
** and conditions see https://www.qt.io/terms-conditions. For further
** information use the contact form at https://www.qt.io/contact-us.
**
** GNU General Public License Usage
** Alternatively, this file may be used under the terms of the GNU
** General Public License version 3 as published by the Free Software
** Foundation with exceptions as appearing in the file LICENSE.GPL3-EXCEPT
** included in the packaging of this file. Please review the following
** information to ensure the GNU General Public License requirements will
** be met: https://www.gnu.org/licenses/gpl-3.0.html.
**
****************************************************************************/

#include "backend.h"
#include "qapplication.h"

#include <QCameraDevice>
#include <QDesktopServices>
#include <QFileInfo>
#include <QJsonArray>
#include <QJsonObject>
#include <QMediaDevices>
#include <QSettings>
#include <QSslSocket>
#include <QVideoSink>

#include "../3rdparty/zxing-cpp/example/ZXingQtReader.h"

using namespace ZXingQt;

Backend::Backend(QObject *parent)
    : QObject(parent)
{
    // This will allow us to open the app with the QR code
    QDesktopServices::setUrlHandler("qtdesignviewer", this, "parseDesignViewerUrl");
    QDesktopServices::setUrlHandler("https", this, "parseDesignViewerUrl");
}

void Backend::initialize()
{
    const QRect screenGeometry = QGuiApplication::primaryScreen()->geometry();
    qDebug() << "Qt Design Viewer";
    qDebug() << "System information:";
    qDebug() << "-- Qt version: " << QT_VERSION_STR;
    qDebug() << "-- OpenSSL support: " << QVariant(QSslSocket::supportsSsl()).toString();
    qDebug() << "-- Screen height: " << QString::number(screenGeometry.height());
    qDebug() << "-- Screen width: " << QString::number(screenGeometry.width());

#ifdef QT_DEBUG
    const QString buildType = "Debug";
#else
    const QString buildType = "Release";
#endif
    const QString buildInfo = QCoreApplication::applicationVersion() + "\nTechnology Preview - "
                              + QString(GIT_VERSION) + "\nQt " + QString(QT_VERSION_STR) + " - "
                              + buildType + " Build" + "\nQt Quick Components "
                              + QString(QT_QUICK_COMPONENTS_VERSION) + "\nOpenSSL support: "
                              + QVariant(QSslSocket::supportsSsl()).toString();
    emit buildInfoChanged(buildInfo);

    QSettings settings;
    m_userHash = settings.value("user/hash").toString();
    if (m_userHash.isEmpty())
        qDebug("User Hash is not registered. Scan QR code to register.");
    else {
        qDebug() << "User Hash: " << m_userHash;
        updateUserProjectList();
    }

    // Check if updateInBackground is enabled
    updateInBackground(settings.value("system/updateInBackground").toBool());
    qDebug() << "Initialization complete";
}

void Backend::updateInBackground(const bool &enabled)
{
    QSettings().setValue("system/updateInBackground", enabled);
    if (enabled) {
        qDebug() << "Update in background is enabled";
        m_backgroundTimer.setInterval(1000 * 10);
        connect(&m_backgroundTimer, &QTimer::timeout, this, [&] {
            qDebug() << "Checking for updates in background";
            if (m_userHash.isEmpty())
                return;

            m_serviceConnector.reset(new ServiceConnector);
            QJsonArray projectList = m_serviceConnector->fetchUserProjectList(m_userHash);
            if (projectList.isEmpty())
                return;

            qDebug() << "New projects available. Updating project list";
            updateUserProjectList();
        });
        m_backgroundTimer.start();
    } else {
        qDebug() << "Update in background is disabled";
        m_backgroundTimer.stop();
    }
}

void Backend::scanQrCode()
{
    m_captureSession.reset(new QMediaCaptureSession(this));
    m_captureSession->setCamera(new QCamera(this));
    m_captureSession->setVideoOutput(new CustomVideoWidget);

    m_captureSession->camera()->setFocusMode(QCamera::FocusModeAuto);

    CustomVideoWidget *videoWidget = qobject_cast<CustomVideoWidget *>(
        m_captureSession->videoOutput());

    connect(m_captureSession->videoSink(),
            &QVideoSink::videoFrameChanged,
            this,
            [&](const QVideoFrame &frame) {
                static int i = 0;
                if (i++ < 30) {
                    return;
                }

                i = 0;

                auto results = ReadBarcodes(frame.toImage());
                for (auto &result : results) {
                    qDebug() << "Text:   " << result.text();
                    qDebug() << "Format: " << result.format();
                    qDebug() << "Content:" << result.contentType();
                    parseDesignViewerUrl(result.text());
                    qobject_cast<CustomVideoWidget *>(m_captureSession->videoOutput())->close();
                }
            });

    connect(videoWidget, &CustomVideoWidget::closed, this, [&] {
        qDebug() << "Video widget closed. Clearing capture session";
        m_captureSession->camera()->stop();
        m_captureSession->camera()->deleteLater();
        m_captureSession->videoOutput()->deleteLater();
        m_captureSession.clear();
    });

    videoWidget->show();
    m_captureSession->camera()->start();
}

void Backend::cacheDemoProjects(const bool &enabled)
{
    QSettings().setValue("system/cacheDemoProjects", enabled);
    if (enabled) {
        qDebug() << "Caching demo projects is enabled";
    } else {
        qDebug() << "Caching demo projects is disabled";
    }
}
void Backend::cacheUserProjects(const bool &enabled)
{
    QSettings().setValue("system/cacheUserProjects", enabled);
    if (enabled) {
        qDebug() << "Caching user projects is enabled";
    } else {
        qDebug() << "Caching user projects is disabled";
    }
}

void Backend::updatePopup(const QString &text, bool indeterminate)
{
    emit popupTextChanged(text);
    emit popupProgressIndeterminateChanged(indeterminate);
    QEventLoop().processEvents(QEventLoop::AllEvents, 1000);
}

void Backend::initializeProjectManager()
{
    if (m_projectManager)
        return;

    m_projectManager.reset(new ProjectManager);
    connect(m_projectManager.data(), &ProjectManager::closingProject, this, [&] {
        emit popupClose();
        m_projectManager.reset();
    });

    m_serviceConnector.reset(new ServiceConnector);
    connect(m_serviceConnector.data(),
            &ServiceConnector::downloadProgress,
            this,
            &Backend::downloadProgress);
}

bool Backend::connectDesignStudio()
{
    if (m_designStudioConnector) {
        qDebug() << "Design Studio Connector is already initialized";
        return true;
    }

    qDebug() << "Initializing Design Studio Connector";

    connect(
        &m_dsConnectorThread,
        &QThread::started,
        this,
        [&] {
            m_designStudioConnector.reset(new DesignStudioConnector);
            connect(m_designStudioConnector.data(),
                    &DesignStudioConnector::networkStatusUpdated,
                    this,
                    &Backend::networkUpdated);

            connect(m_designStudioConnector.data(),
                    &DesignStudioConnector::projectReceived,
                    this,
                    [this](const QByteArray &projectData) {
                        qDebug() << "Project received from Design Studio";
                        initializeProjectManager();
                        emit popupOpen();
                        updatePopup("Unpacking project...");
                        qDebug() << "Project data size: " << projectData.size();
                        const QString projectPath = m_projectManager->unpackProject(projectData);

                        if (projectPath.isEmpty()) {
                            qCritical()
                                << "Could not unpack project. Please check the logs for more "
                                   "information.";
                            emit popupClose();
                            return;
                        }

                        qDebug() << "Project unpacked to " << projectPath;
                        updatePopup("Running project...");

                        if (!m_projectManager->runProject(projectPath)) {
                            qCritical() << "Could not run project. Please check the logs for more "
                                           "information.";
                        } else {
                            m_projectManager->showAppWindow();
                        }
                        emit popupClose();
                    });
            if (!m_designStudioConnector->initialize()) {
                qCritical()
                    << "Could initialize server. Please check the logs for more information.";
                m_designStudioConnector.reset();
            }
        },
        Qt::DirectConnection);

    m_dsConnectorThread.start();

    qDebug() << "Design Studio Connector is initialized";
    return true;
}

void Backend::disconnectDesignStudio()
{
    if (!m_designStudioConnector) {
        qDebug() << "Design Studio Connector is not initialized";
        return;
    }

    qDebug() << "Disconnecting from Design Studio";

    connect(
        &m_dsConnectorThread,
        &QThread::finished,
        this,
        [&] {
            qDebug() << "Design Studio Connector is disconnected";
            m_designStudioConnector.reset();
        },
        Qt::DirectConnection);

    m_dsConnectorThread.quit();
    m_dsConnectorThread.wait();
}

void Backend::runDemoProject(const QString &projectName)
{
    initializeProjectManager();
    qDebug() << "Checking if demo project is cached for " << projectName;
    emit popupOpen();

    // sample project info
    // [{"lastUpdate":1701947766739.9812,"name":"ClusterTutorial.qmlrc"}]
    const QJsonArray projectList = m_serviceConnector->fetchUserDemoList();
    QJsonObject projectInfo;
    for (auto project : projectList) {
        if (projectName == project.toObject().value("name").toString().remove(".qmlrc")) {
            projectInfo = project.toObject();
            break;
        }
    }

    const bool cached = m_projectManager->isDemoProjectCached(projectInfo);

    if (!cached) {
        updatePopup("Downloading demo project...", false);

        const QString url = "https://designviewer.qt.io/qmlprojects/demos/" + projectName
                            + ".qmlrc";
        QByteArray project = m_serviceConnector->fetchProject(url);
        qDebug() << "Demo project is not cached. Trying to download from " << url << " ...";

        updatePopup("Caching demo project...");
        if (!m_projectManager->cacheDemoProject(project, projectInfo)) {
            qCritical()
                << "Could not cache demo project. Please check the logs for more information.";
            emit popupClose();
            return;
        }
    } else {
        qDebug() << "Demo project is cached. Running cached project...";
    }

    updatePopup("Running demo project...");
    if (!m_projectManager->runDemoProject(projectName))
        qCritical() << "Could not run demo project. Please check the logs for more information.";
    else {
        updatePopup("Showing the app window...");
        m_projectManager->showAppWindow();
    }

    emit popupClose();
}

void Backend::clearDemoCaches()
{
    emit popupOpen();
    updatePopup("Clearing demo caches...");
    m_projectManager.reset(new ProjectManager);
    m_projectManager->clearDemoCaches();
    m_projectManager.reset();
    emit popupClose();
}

void Backend::runUserProject(const QString &projectName)
{
    initializeProjectManager();
    updatePopup("Running user project");
    emit popupOpen();

    qDebug() << "Running user project:" << projectName;

    qDebug("Checking if project is cached. Getting list of available projects...");
    const QJsonArray projectList = m_serviceConnector->fetchUserProjectList(m_userHash);

    QString projectLastModified;
    QString projectId;
    QJsonObject projectInfo;
    for (auto project : projectList) {
        if (projectName == project.toObject().value("appName").toString()) {
            projectInfo = project.toObject();
            break;
        }
    }

    const bool projectCached = m_projectManager->isProjectCached(projectInfo);

    if (!projectCached) {
        qDebug("Project is not cached. Downloading...");
        updatePopup("Project is not cached. Downloading...", false);
        QByteArray projectData = m_serviceConnector->fetchUserProject(m_userHash, projectName);

        updatePopup("Caching user project...");
        if (!m_projectManager->cacheProject(projectData, projectInfo)) {
            qCritical() << "Could not cache project. Please check the logs for more information.";
            emit popupClose();
            return;
        }
    }

    qDebug("Project is cached. Running cached project...");
    updatePopup("Running cached project...");
    if (!m_projectManager->runCachedProject(projectInfo))
        qCritical() << "Could not run project. Please check the logs for more information.";
    else {
        updatePopup("Showing the app window...");
        m_projectManager->showAppWindow();
    }

    emit popupClose();
}

void Backend::runOnlineProject(const QString &url)
{
    initializeProjectManager();
    emit popupOpen();
    updatePopup("Downloading...", false);
    QByteArray projectData = m_serviceConnector->fetchProject(url);
    if (projectData.isEmpty()) {
        qCritical() << "Could not download project. Please check the logs for more information.";
        return;
    }

    updatePopup("Unpacking project...");
    QString projectPath = m_projectManager->unpackProject(projectData);
    updatePopup("Running project...");

    if (!m_projectManager->runProject(projectPath))
        qCritical() << "Could not run project. Please check the logs for more information.";
    else {
        m_projectManager->showAppWindow();
    }

    emit popupClose();
}

void Backend::registerUser(const QUrl &url)
{
    const QString userHash = url.toString().remove("qtdesignviewer://");
    qDebug() << "Registering User Hash: " << userHash;
    m_userHash = userHash;
    QSettings().setValue("user/hash", m_userHash);
    emit userRegistered();
    updateUserProjectList();
}

void Backend::updateUserProjectList()
{
    qDebug() << "Fetching available project list for user: " << m_userHash;
    m_serviceConnector.reset(new ServiceConnector);
    QJsonArray projectList = m_serviceConnector->fetchUserProjectList(m_userHash);

    if (projectList.isEmpty()) {
        qCritical("Could not fetch available project list");
        return;
    }

    if (projectList == m_projectListArray) {
        qDebug("No new projects available");
        return;
    }
    m_projectListArray = projectList;
    m_projectList.clear();
    qDebug("List of available projects fetched:");
    for (const auto &project : projectList) {
        const QString projectName{project.toObject().value("appName").toString()};
        qDebug() << "--" << projectName;
        m_projectList << projectName;
    }
    emit projectListChanged();
}

void Backend::parseDesignViewerUrl(const QUrl &url)
{
    QString urlData = url.toString().remove("qtdesignviewer://");
    // urlData could be either a direct url to the project or a user hash
    // If it is a user hash, we register the user and fetch the project list
    // If it is a project url, we submit the url to the text field
    // sample url: qtdesignviewer://https://<url>/<project_name>.qmlrc
    // sample user hash: qtdesignviewer://17e8907b3b84029384hs8djshdu38476
    if (urlData.isEmpty())
        return;
    else if (urlData.startsWith("https//")) {
        urlData.replace("https//", "https://");
        emit urlUpdated(urlData);
    } else if (urlData.startsWith("https://")) {
        emit urlUpdated(urlData);
    } else {
        qDebug() << "Registering user from QR code";
        registerUser(url);
    }
}
