/****************************************************************************
**
** Copyright (C) 2023 The Qt Company Ltd.
** Contact: https://www.qt.io/licensing/
**
** This file is part of the Qt Design Viewer of the Qt Toolkit.
**
** Commercial License Usage
** Licensees holding valid commercial Qt licenses may use this file in
** accordance with the commercial license agreement provided with the
** Software or, alternatively, in accordance with the terms contained in
** a written agreement between you and The Qt Company. For licensing terms
** and conditions see https://www.qt.io/terms-conditions. For further
** information use the contact form at https://www.qt.io/contact-us.
**
** GNU General Public License Usage
** Alternatively, this file may be used under the terms of the GNU
** General Public License version 3 as published by the Free Software
** Foundation with exceptions as appearing in the file LICENSE.GPL3-EXCEPT
** included in the packaging of this file. Please review the following
** information to ensure the GNU General Public License requirements will
** be met: https://www.gnu.org/licenses/gpl-3.0.html.
**
****************************************************************************/

#include "dsConnector.h"

#include <QNetworkInterface>

void DesignStudioConnector::receiveProject()
{
    QByteArray data = m_tcpSocket->readAll();
    // qDebug() << "Data from Design Studio:" << data;
    if (data.startsWith("qres")) {
        qDebug() << "TCP:: Received project start delimeter";
        m_projectData.clear();
        m_projectData.append(data);
        m_receivingData = true;
    } else if (data.contains("::qmlrc-end::")) {
        qDebug() << "TCP:: Received project end delimeter";
        // qDebug() << "TCP:: Last data sequence: " << data;
        m_projectData.append(data.mid(0, data.size() - 13));
        emit projectReceived(m_projectData);
        m_receivingData = false;
    } else if (m_receivingData) {
        qDebug() << "TCP:: Received data sequence";
        m_projectData.append(data);
    } else {
        qDebug() << "TCP:: Received unknown data:" << data;
    }
}

bool DesignStudioConnector::initTcpServer()
{
    const bool retVal = m_tcpServer.listen(QHostAddress::Any, m_tcpPort);

    if (!retVal) {
        qDebug() << "Failed to listen on port " << m_tcpPort;
        emit networkStatusUpdated("Failed to bind on port " + QString::number(m_tcpPort));
        return false;
    }

    qDebug() << "Listening on port " << m_tcpPort;
    connect(&m_tcpServer, &QTcpServer::newConnection, this, [this]() {
        qDebug() << "New connection from Design Studio";
        emit networkStatusUpdated("Design Studio is connected.\n Waiting for project...");
        m_tcpSocket.reset(m_tcpServer.nextPendingConnection());
        m_broadcastTimer.stop();

        connect(m_tcpSocket.data(), &QTcpSocket::disconnected, this, [this]() {
            qDebug() << "Disconnected from Design Studio";
            emit networkStatusUpdated("\nLocal IP: " + m_ipv4Addr
                                      + "\nWaiting for Design Studio to connect...");
            m_broadcastTimer.start();
            m_projectData.clear();
        });

        connect(m_tcpSocket.data(),
                &QTcpSocket::readyRead,
                this,
                &DesignStudioConnector::receiveProject);
    });

    return true;
}

void DesignStudioConnector::initBroadcast()
{
    // get ipv4 address
    const QList<QHostAddress> list = QNetworkInterface::allAddresses();
    for (const QHostAddress &address : list) {
        if (address.protocol() == QAbstractSocket::IPv4Protocol
            && address != QHostAddress::LocalHost) {
            qDebug() << "Found IPv4 address:" << address.toString();
            m_ipv4Addr = address.toString();
            break;
        }
    }

    qDebug() << "Advertising from" << m_ipv4Addr;
    emit networkStatusUpdated("\nLocal IP: " + m_ipv4Addr
                              + "\nWaiting for Design Studio to connect...");

    // Broadcast a message to the network in each 2 seconds
    // to let the Design Studio know that we are here
    connect(&m_broadcastTimer, &QTimer::timeout, this, [this]() {
        QByteArray datagram = "qt_ui_viewer_at:" + m_ipv4Addr.toUtf8() + ":"
                              + QByteArray::number(m_tcpPort);

        m_udpSocket.writeDatagram(datagram.data(),
                                  datagram.size(),
                                  QHostAddress::Broadcast,
                                  m_udpPort);

        qDebug() << "Broadcasting datagram:" << datagram;
    });

    m_broadcastTimer.setSingleShot(false);
    m_broadcastTimer.start(2000);
}

bool DesignStudioConnector::initialize()
{
    if (!initTcpServer())
        return false;

    initBroadcast();
    return true;
}
